@@ -9,7 +9,7 @@ use strict;
 use warnings;
 
 package Net::SIP::StatelessProxy;
-use fields qw( dispatcher rewrite_contact nathelper force_rewrite );
+use fields qw( dispatcher rewrite_contact nathelper force_rewrite respcode );
 
 use Net::SIP::Util ':all';
 use Digest::MD5 qw(md5);
@@ -50,6 +50,7 @@ sub new {
     };
     $self->{nathelper} = delete $args{nathelper};
     $self->{force_rewrite} = delete $args{force_rewrite};
+    $self->{respcode} = [ {},{} ];
 
     return $self;
 }
@@ -641,6 +642,11 @@ sub do_nat {
 	: ( undef,$packet )
 	;
     my $method = $request ? $request->method : '';
+    my $track_resp_code;
+    if ($response and $response->method eq 'INVITE') {
+	my $code = $response->code;
+	$track_resp_code = $code if $code>=400;
+    }
 
     # NAT for anything with SDP body
     # activation and close of session will be done on ACK|CANCEL|BYE
@@ -649,7 +655,7 @@ sub do_nat {
 	or $method eq 'CANCEL'
 	or $method eq 'BYE' ) {
 	DEBUG( 100, "no NAT because no SDP body and method is $method" );
-	return;
+	return if ! $track_resp_code;
     }
 
 
@@ -659,21 +665,17 @@ sub do_nat {
     # Leg, delimited by "\0"
     my ($idfrom,$idto);
 
-    if ( my $from = $packet->get_header( 'from' ) ) {
-	my ($data,$param) = sip_hdrval2parts( from => $from );
-	my $tag = $param->{tag} || '';
-	$idfrom = "$data\0$tag";
-    } else {
-	return [ 0,'no FROM header in packet' ]
+    for([from => \$idfrom], [to => \$idto]) {
+	my ($k,$idref) = @$_;
+	if (my $v = $packet->get_header($k) ) {
+	    my ($uri,$param) = sip_hdrval2parts(from => $v);
+	    my ($dom,$user,$proto) = sip_uri2parts($uri);
+	    $$idref = "$proto:$user\@$dom\0".($param->{tag} || '');
+	} else {
+	    return [ 0,'no '.uc($k).' header in packet' ]
+	}
     }
 
-    if ( my $to = $packet->get_header( 'to' ) ) {
-	my ($data,$param) = sip_hdrval2parts( from => $to );
-	my $tag = $param->{tag} || '';
-	$idto = "$data\0$tag";
-    } else {
-	return [ 0,'no TO header in packet' ]
-    }
 
     # side is either 0 (request) or 1 (response)
     # If a request comes in 'from' points to the incoming_leg while
@@ -697,6 +699,18 @@ sub do_nat {
 	or return [ 0,'no CSEQ in packet' ];
     my $callid = $packet->callid;
 
+    if ($track_resp_code) {
+	my $rc = $self->{respcode}[0];
+	if (keys(%$rc)>5000) {
+	    # expire entries
+	    $self->{respcode}[1] = $rc;
+	    $rc = $self->{respcode}[0] = {};
+	}
+	$rc->{$callid,$cseq,$idfrom,$idto} = $track_resp_code;
+	# no NAT to do, we just needed to track the response code
+	return;
+    }
+
     # CANCEL|BYE will be handled first to close session
     # no NAT will be done, even if the packet contains SDP (which makes no sense)
     if ( $method eq 'CANCEL' ) {
@@ -733,8 +747,16 @@ sub do_nat {
 
     if ( ! $nathelper->activate_session( $callid,$cseq,$idfrom,$idto ) ) {
 	if ( $method eq 'ACK' ) {
-	    DEBUG( 50,"session $callid|$cseq $idfrom -> $idto still incomplete in ACK" );
-	    return [ 0,'incomplete session in ACK' ]
+	    my $code = $self->{respcode}[0]{$callid,$cseq,$idfrom,$idto}
+		|| $self->{respcode}[1]{$callid,$cseq,$idfrom,$idto}
+		|| -1;
+	    if ($code < 400) {
+		DEBUG( 50,"session $callid|$cseq $idfrom -> $idto still incomplete in ACK" );
+		return [ 0,'incomplete session in ACK' ]
+	    } else {
+		# ignore problem, ACK to response with error code
+		DEBUG( 100, "session $callid|$cseq $idfrom -> ACK to failure response" );
+	    }
 	} else {
 	    # ignore problem, session not yet complete
 	    DEBUG( 100, "session $callid|$cseq $idfrom -> $idto not yet complete" );
